# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/7 21:07
"""
import pandas as pd
from tqdm import tqdm
from copy import deepcopy
from loguru import logger
from typing import List
from collections import OrderedDict
from czsc import signals
from czsc.traders.base import CzscSignals, BarGenerator, CzscTrader
from czsc.traders.sig_parse import get_signals_config, get_signals_freqs
from czsc.objects import Signal, Factor, Event, Operate, Position
from test.test_analyze import read_1min, read_daily


def test_object_position():
    bars = read_daily()
    bg = BarGenerator(base_freq='日线', freqs=['周线', '月线'])
    for bar in bars[:1000]:
        bg.update(bar)

    opens = [
        Event(name='开多', operate=Operate.LO, factors=[
            Factor(name="站上SMA5", signals_all=[
                Signal("日线_D1B_BUY1_一买_任意_任意_0"),
            ])
        ]),
        Event(name='开空', operate=Operate.SO, factors=[
            Factor(name="跌破SMA5", signals_all=[
                Signal("日线_D1B_BUY1_一卖_任意_任意_0"),
            ])
        ]),
    ]

    # 没有出场条件的测试
    pos = Position(name="测试A", symbol=bg.symbol, opens=opens, exits=[], interval=0, timeout=20, stop_loss=300)

    signals_config = get_signals_config(pos.unique_signals)
    cs = CzscSignals(deepcopy(bg), signals_config=signals_config)
    for bar in bars[1000:]:
        cs.update_signals(bar)
        pos.update(cs.s)

    df = pd.DataFrame(pos.pairs)
    assert df.shape == (16, 11)
    assert len(cs.s) == 12

    exits = [
        Event(name='平多', operate=Operate.LE, factors=[
            Factor(name="跌破SMA5", signals_all=[
                Signal("日线_D0停顿分型_BE辅助V230106_看空_强_任意_0"),
            ])
        ]),
        Event(name='平空', operate=Operate.SE, factors=[
            Factor(name="站上SMA5", signals_all=[
                Signal("日线_D0停顿分型_BE辅助V230106_看多_强_任意_0"),
            ])
        ]),
    ]

    pos = Position(name="测试B", symbol=bg.symbol, opens=opens, exits=exits, interval=0, timeout=20, stop_loss=300)

    assert len(pos.unique_signals) == 4
    assert len(pos.events[0].unique_signals) == 1

    signals_config = get_signals_config(pos.unique_signals)
    cs = CzscSignals(deepcopy(bg), signals_config=signals_config)
    for bar in bars[1000:]:
        cs.update_signals(bar)
        pos.update(cs.s)

    df = pd.DataFrame(pos.pairs)
    assert df.shape == (17, 11)
    assert len(cs.s) == 13

    # 测试 dump 和 load
    pos_x = pos.dump()
    assert isinstance(pos_x, dict)
    pos_x['name'] = "测试C"
    pos_y = Position.load(pos_x)
    cs = CzscSignals(deepcopy(bg), signals_config=get_signals_config(pos_y.unique_signals))
    for bar in bars[1000:]:
        cs.update_signals(bar)
        pos_y.update(cs.s)

    assert pos_y.name == "测试C"
    df = pd.DataFrame(pos_y.pairs)
    assert df.shape == (17, 11)
    assert len(cs.s) == 13


def test_generate_czsc_signals():
    from czsc.traders.base import generate_czsc_signals

    bars = read_daily()
    signals_seq = [
        "日线_D1B_BUY1_一买_任意_任意_0",
        "日线_D1B_BUY1_一卖_任意_任意_0",
        "日线_D0停顿分型_BE辅助V230106_看空_强_任意_0"
    ]

    signals_config = get_signals_config(signals_seq)

    # 通过 signals_seq 得到 freqs
    freqs = get_signals_freqs(signals_seq)

    # 通过 signals_config 得到 freqs
    freqs1 = get_signals_freqs(signals_config)

    assert freqs == freqs1
    res = generate_czsc_signals(bars, signals_config=signals_config, freqs=freqs, sdt="20100101", init_n=500)
    rdf = generate_czsc_signals(bars, signals_config=signals_config, freqs=freqs, sdt="20100101", init_n=500, df=True)

    assert len(res) == len(rdf)


def test_czsc_trader():
    bars = read_daily()

    sdt = "20100101"
    init_n = 2000
    sdt = pd.to_datetime(sdt)
    bars_left = [x for x in bars if x.dt < sdt]
    if len(bars_left) <= init_n:
        bars_left = bars[:init_n]
        bars_right = bars[init_n:]
    else:
        bars_right = [x for x in bars if x.dt >= sdt]

    bg = BarGenerator(base_freq='日线', freqs=['周线', '月线'])
    for bar in bars_left:
        bg.update(bar)

    def __create_sma5_pos():
        opens = [
            Event(name='开多', operate=Operate.LO, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_D1B_BUY1_一买_任意_任意_0"),
                ])
            ]),
            Event(name='开空', operate=Operate.SO, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_D1B_BUY1_一卖_任意_任意_0"),
                ])
            ]),
        ]

        exits = [
            Event(name='平多', operate=Operate.LE, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_D0停顿分型_BE辅助V230106_看空_强_任意_0"),
                ])
            ]),
            Event(name='平空', operate=Operate.SE, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_D0停顿分型_BE辅助V230106_看多_强_任意_0"),
                ])
            ]),
        ]

        pos = Position(symbol=bg.symbol, opens=opens, exits=exits, interval=0, timeout=20, stop_loss=100, name="测试A")

        return pos

    def __create_sma10_pos():
        opens = [
            Event(name='开多', operate=Operate.LO, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_D2B_BUY1_一买_任意_任意_0"),
                ])
            ]),
            Event(name='开空', operate=Operate.SO, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_D2B_BUY1_一卖_任意_任意_0"),
                ])
            ]),
        ]

        exits = [
            Event(name='平多', operate=Operate.LE, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_D0停顿分型_BE辅助V230106_看空_强_任意_0"),
                ])
            ]),
            Event(name='平空', operate=Operate.SE, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_D0停顿分型_BE辅助V230106_看多_强_任意_0"),
                ])
            ]),
        ]

        pos = Position(symbol=bg.symbol, opens=opens, exits=exits, interval=0, timeout=20, stop_loss=100, name="测试B")
        return pos

    def __create_sma20_pos():
        opens = [
            Event(name='开多', operate=Operate.LO, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_D3B_BUY1_一买_任意_任意_0"),
                ])
            ]),
            Event(name='开空', operate=Operate.SO, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_D3B_BUY1_一卖_任意_任意_0"),
                ])
            ]),
        ]

        exits = [
            Event(name='平多', operate=Operate.LE, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_D0停顿分型_BE辅助V230106_看空_强_任意_0"),
                ])
            ]),
            Event(name='平空', operate=Operate.SE, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_D0停顿分型_BE辅助V230106_看多_强_任意_0"),
                ])
            ]),
        ]

        pos = Position(symbol=bg.symbol, opens=opens, exits=exits, interval=0, timeout=20, stop_loss=100, name="测试C")
        return pos

    positions = [__create_sma5_pos(), __create_sma10_pos(), __create_sma20_pos()]
    signals_seq = []
    for _pos in positions:
        signals_seq.extend(_pos.unique_signals)

    signals_config = get_signals_config(list(set(signals_seq)))
    # 通过 update 执行
    ct = CzscTrader(deepcopy(bg), signals_config=signals_config,
                    positions=[__create_sma5_pos(), __create_sma10_pos(), __create_sma20_pos()])
    for bar in bars_right:
        ct.update(bar)
        for _pos in ct.positions:
            if _pos.pos_changed:
                assert _pos.operates[-1]['dt'] == _pos.end_dt
                print(_pos.name, _pos.operates[-1], _pos.end_dt, _pos.pos)
                assert ct.pos_changed

    assert list(ct.positions[0].dump(False).keys()) == ['symbol', 'name', 'opens', 'exits', 'interval', 'timeout',
                                                        'stop_loss', 'T0']
    assert list(ct.positions[0].dump(True).keys()) == ['symbol', 'name', 'opens', 'exits', 'interval', 'timeout',
                                                       'stop_loss', 'T0', 'pairs', 'holds']
    assert [x.pos for x in ct.positions] == [0, 0, 0]

    # 测试自定义仓位集成
    def _weighted_ensemble(positions: List[Position]):
        return 0.5 * positions[0].pos + 0.5 * positions[1].pos

    assert ct.get_ensemble_pos(_weighted_ensemble) == 0
    assert ct.get_ensemble_pos('vote') == 0
    assert ct.get_ensemble_pos('max') == 0
    assert ct.get_ensemble_pos('mean') == 0

    # 通过 on_bar 执行
    ct1 = CzscTrader(deepcopy(bg), signals_config=signals_config,
                     positions=[__create_sma5_pos(), __create_sma10_pos(), __create_sma20_pos()])
    for bar in bars_right:
        ct1.on_bar(bar)
        # print(ct1.s)
        print(
            f"{ct1.end_dt}: pos_seq = {[x.pos for x in ct1.positions]}mean_pos = {ct1.get_ensemble_pos('mean')}; vote_pos = {ct1.get_ensemble_pos('vote')}; max_pos = {ct1.get_ensemble_pos('max')}")

    assert [x.pos for x in ct1.positions] == [0, 0, 0]

    assert len(ct1.positions[0].pairs) == len(ct.positions[0].pairs)
    assert len(ct1.positions[1].pairs) == len(ct.positions[1].pairs)
    assert len(ct1.positions[2].pairs) == len(ct.positions[2].pairs)

    # 通过 on_sig 执行
    from czsc.traders.base import generate_czsc_signals
    res = generate_czsc_signals(bars, signals_config=signals_config, freqs=['周线', '月线'], sdt=sdt, init_n=init_n)
    ct2 = CzscTrader(positions=[__create_sma5_pos(), __create_sma10_pos(), __create_sma20_pos()])
    for sig in res:
        ct2.on_sig(sig)
        print(
            f"{ct2.end_dt}: pos_seq = {[x.pos for x in ct2.positions]}mean_pos = {ct2.get_ensemble_pos('mean')}; vote_pos = {ct2.get_ensemble_pos('vote')}; max_pos = {ct2.get_ensemble_pos('max')}")

    assert [x.pos for x in ct2.positions] == [0, 0, 0]

    assert len(ct1.positions[0].pairs) == len(ct2.positions[0].pairs)
    assert len(ct1.positions[1].pairs) == len(ct2.positions[1].pairs)
    assert len(ct1.positions[2].pairs) == len(ct2.positions[2].pairs)




# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2019/10/29 15:01
"""
from czsc import envs
from czsc import ai
from czsc import fsa
from czsc import utils
from czsc import traders
from czsc import sensors
from czsc import aphorism
from czsc.analyze import CZSC
from czsc.objects import Freq, Operate, Direction, Signal, Factor, Event, RawBar, NewBar, Position
from czsc.utils.cache import home_path, get_dir_size, empty_cache_path
from czsc.traders import CzscTrader, CzscSignals, generate_czsc_signals, check_signals_acc, get_unique_signals
from czsc.traders import PairsPerformance, combine_holds_and_pairs, combine_dates_and_pairs, stock_holds_performance
from czsc.traders import DummyBacktest, SignalsParser, get_signals_by_conf, get_signals_config, get_signals_freqs
from czsc.strategies import CzscStrategyBase
from czsc.utils import KlineChart, BarGenerator, resample_bars, dill_dump, dill_load, read_json, save_json
from czsc.utils import get_sub_elements, get_py_namespace, freqs_sorted, x_round, import_by_name, create_grid_params
from czsc.utils import cal_trade_price, cross_sectional_ic
from czsc.sensors import holds_concepts_effect, StocksDaySensor, ThsConceptsSensor, SignalsPerformance


__version__ = "0.9.16"
__author__ = "zengbin93"
__email__ = "zeng_bin8888@163.com"
__date__ = "20230404"


def welcome():
    print(f"欢迎使用CZSC！当前版本标识为 {__version__}@{__date__}\n")
    aphorism.print_one()

    print(f"CZSC环境变量："
          f"czsc_min_bi_len = {envs.get_min_bi_len()}; "
          f"czsc_max_bi_num = {envs.get_max_bi_num()}; "
          f"czsc_bi_change_th = {envs.get_bi_change_th()}")


if envs.get_welcome():
    welcome()


if get_dir_size(home_path) > pow(1024, 3) and envs.get_verbose():
    print(f"{home_path} 目录缓存超过1GB，请适当清理。调用 czsc.empty_cache_path 可以直接清空缓存")



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/3/29 10:04
describe: 
"""
import re
from loguru import logger
from parse import parse
from difflib import SequenceMatcher
from czsc.objects import Signal
from czsc.utils import import_by_name


class SignalsParserBackup:
    """解析一串信号，生成信号函数配置"""

    def __init__(self, signals_module='czsc.signals', **kwargs):
        """

        :param signals_module: 指定信号函数所在模块
        :param kwargs:
            usr_parse_map: 用户自定义信号函数解析方法，字典类型，key 为信号函数名，value 为解析方法
        """
        self.signals_module = signals_module
        sig_name_map = {}
        sig_pats_map = {}

        signals_module = import_by_name(signals_module)
        for name in dir(signals_module):
            if "_" not in name:
                continue

            try:
                doc = getattr(signals_module, name).__doc__
                # 解析信号函数参数
                pats = re.findall(r"参数模板：\"(.*)\"", doc)
                if pats:
                    sig_pats_map[name] = pats[0]

                # 解析信号列表
                sigs = re.findall(r"Signal\('(.*)'\)", doc)
                if sigs:
                    sig_name_map[name] = [Signal(x) for x in sigs]

            except Exception as e:
                logger.error(f"解析信号函数 {name} 出错：{e}")

        self.sig_name_map = sig_name_map
        self.sig_pats_map = sig_pats_map

        # 自动获取解析函数
        self._parse_map = {k: getattr(self, f"_SignalsParser__parse_{k}") for k in self.sig_name_map.keys()
                           if getattr(self, f"_SignalsParser__parse_{k}", None)}

        # 用户自定义信号函数解析方法传入
        if kwargs.get("usr_parse_map", None):
            self._parse_map.update(kwargs.get("usr_parse_map"))

    def parse_params(self, name, signal):
        """获取信号函数参数

        :param name: 信号函数名称
        :param signal: 需要解析的信号
        :return:
        """
        key = Signal(signal).key
        pats = self.sig_pats_map.get(name, None)
        if not pats:
            return None

        try:
            params = parse(pats, key).named
            if 'di' in params:
                params['di'] = int(params['di'])

            params['name'] = f"{self.signals_module}.{name}"
            return params
        except Exception as e:
            logger.error(f"解析信号 {signal} - {name} - {pats} 出错：{e}")
            return None

    def get_function_name(self, signal):
        """获取信号函数名称"""
        sig_name_map = self.sig_name_map
        _signal = Signal(signal)
        _k3_match = list({k for k, v in sig_name_map.items() if v[0].k3 == _signal.k3})
        # 优先匹配 k3，满足条件直接返回
        if len(_k3_match) == 1:
            return _k3_match[0]

        if len(_k3_match) > 1:
            logger.warning(f"信号 {signal} 有多个匹配函数：{_k3_match}，请手动解析信号")
            return None

        _signal_k2_k3 = _signal.k2 + _signal.k3
        scores = {}
        for k, v in sig_name_map.items():
            # 计算 k2, k3 的相似度
            _vs = [SequenceMatcher(None, s.k2 + s.k3, _signal_k2_k3).ratio() for s in v]
            if max(_vs) >= 0.8:
                scores[k] = max(_vs)

        if not scores:
            return None

        return max(scores, key=scores.get)

    def parse(self, signal_seq):
        """解析信号序列"""
        res = []
        for signal in signal_seq:
            name = self.get_function_name(signal)

            # 首先使用参数模板进行解析
            if name in self.sig_pats_map:
                row = self.parse_params(name, signal)
                if row and row not in res:
                    res.append(row)

            # 其次使用信号函数名称对应的解析方法进行解析
            elif name in self._parse_map:
                row = self._parse_map[name](signal)
                row['name'] = f"{self.signals_module}.{name}"
                if row not in res:
                    res.append(row)

            else:
                logger.warning(f"未找到解析函数：{name}，请手动解析信号：{signal}")

        return res

    def parse_with_name(self, signal_map):
        """解析信号字典，信号字典的 key 为信号函数名，value 为信号序列"""
        res = []
        for name, signal_seq in signal_map.items():
            if name in self._parse_map:
                for signal in signal_seq:
                    row = self._parse_map[name](signal)
                    row['name'] = f"czsc.signals.{name}"
                    if row not in res:
                        res.append(row)
            else:
                logger.warning(f"未找到解析函数：{name}，请手动解析信号：{signal_seq}")
        return res

    @staticmethod
    def __remove_duplicates(_res):
        # 去除重复的信号配置
        _res = [dict(t) for t in {tuple(d.items()) for d in _res}]
        return _res

    @staticmethod
    def __parse_bar_single_V230214(signal):
        # https://czsc.readthedocs.io/en/0.9.13/api/czsc.signals.bar_single_V230214.html
        pats = re.findall(r"(.*?)_D(\d+)T(\d+)_", signal)[0]
        _row = {"freq": pats[0], "di": int(pats[1]), 't': int(pats[2]) / 10}
        return _row

    @staticmethod
    def __parse_cxt_third_bs_V230319(signal):
        pats = re.findall(r"(.*?)_D(\d+)(\D+)(\d+)_", signal)[0]
        _row = {"freq": pats[0], "di": int(pats[1]), 'ma_type': pats[2], 'timeperiod': int(pats[3])}
        return _row

    @staticmethod
    def __parse_byi_bi_end_V230107(signal):
        pats = re.findall(r"(.*?)_", signal)
        _row = {"freq": pats[0]}
        return _row

    @staticmethod
    def __parse_byi_bi_end_V230106(signal):
        pats = re.findall(r"(.*?)_", signal)
        _row = {"freq": pats[0]}
        return _row

    @staticmethod
    def __parse_bar_accelerate_V221110(signal):
        # https://czsc.readthedocs.io/en/0.9.13/api/czsc.signals.bar_accelerate_V221110.html
        pats = re.findall(r"(.*?)_D(\d+)W(\d+)_", signal)[0]
        _row = {"freq": pats[0], "di": int(pats[1]), 'window': int(pats[2])}
        return _row

    @staticmethod
    def __parse_bar_accelerate_V221118(signal):
        # https://czsc.readthedocs.io/en/0.9.13/api/czsc.signals.bar_accelerate_V221118.html
        pats = re.findall(r"(.*?)_D(\d+)W(\d+)(\D+)(\d+)_", signal)[0]
        _row = {
            "freq": pats[0],
            "di": int(pats[1]),
            "window ": int(pats[2]),
            "ma_type": pats[3],
            "timeperiod ": int(pats[4]),
        }
        return _row

    @staticmethod
    def __parse_bar_bpm_V230227(signal):
        # https://czsc.readthedocs.io/en/0.9.13/api/czsc.signals.bar_bpm_V230227.html
        pats = re.findall(r"(.*?)_D(\d+)N(\d+)T(\d+)_", signal)[0]
        _row = {"freq": pats[0], "di": int(pats[1]), "n ": int(pats[2]), "th ": int(pats[3])}
        return _row


class SignalsParser:
    """解析一串信号，生成信号函数配置"""

    def __init__(self, signals_module='czsc.signals', **kwargs):
        """

        :param signals_module: 指定信号函数所在模块
        :param kwargs:
            usr_parse_map: 用户自定义信号函数解析方法，字典类型，key 为信号函数名，value 为解析方法
        """
        self.signals_module = signals_module
        sig_name_map = {}
        sig_pats_map = {}

        signals_module = import_by_name(signals_module)
        for name in dir(signals_module):
            if "_" not in name:
                continue

            try:
                doc = getattr(signals_module, name).__doc__
                # 解析信号函数参数
                pats = re.findall(r"参数模板：\"(.*)\"", doc)
                if pats:
                    sig_pats_map[name] = pats[0]

                # 解析信号列表
                sigs = re.findall(r"Signal\('(.*)'\)", doc)
                if sigs:
                    sig_name_map[name] = [Signal(x) for x in sigs]

            except Exception as e:
                logger.error(f"解析信号函数 {name} 出错：{e}")

        self.sig_name_map = sig_name_map
        self.sig_pats_map = sig_pats_map

    def parse_params(self, name, signal):
        """获取信号函数参数

        :param name: 信号函数名称
        :param signal: 需要解析的信号
        :return:
        """
        key = Signal(signal).key
        pats = self.sig_pats_map.get(name, None)
        if not pats:
            return None

        try:
            params = parse(pats, key).named
            if 'di' in params:
                params['di'] = int(params['di'])

            params['name'] = f"{self.signals_module}.{name}"
            return params
        except Exception as e:
            logger.error(f"解析信号 {signal} - {name} - {pats} 出错：{e}")
            return None

    def get_function_name(self, signal):
        """获取信号函数名称"""
        sig_name_map = self.sig_name_map
        _signal = Signal(signal)
        _k3_match = list({k for k, v in sig_name_map.items() if v[0].k3 == _signal.k3})
        # 优先匹配 k3，满足条件直接返回
        if len(_k3_match) == 1:
            return _k3_match[0]
        else:
            logger.error(f"信号 {signal} 有多个匹配函数：{_k3_match}，请手动解析信号")
            return None

    def parse(self, signal_seq):
        """解析信号序列"""
        res = []
        for signal in signal_seq:
            name = self.get_function_name(signal)
            if name in self.sig_pats_map:
                row = self.parse_params(name, signal)
                if row and row not in res:
                    res.append(row)
            else:
                logger.warning(f"未找到解析函数：{name}，请手动解析信号：{signal}")
        return res
